#include "file_tab.h"
#include "threading.h"

#include <xxhash/xxhash.h>

#define NAME_CAP(cap) ((cap)*128)
typedef struct {
    vy_file_id *ids;
    unsigned int *name_offsets;
    char *names;
    unsigned int capacity;
    unsigned int name_head;

    vy_mutex *mutex;
} vy_file_tab;

static vy_file_tab _file_tab;

vy_result vyInitFileTab(unsigned int max_files) {
    _file_tab.ids = calloc(max_files, sizeof(vy_file_id));
    if (!_file_tab.ids)
        return 1;
    _file_tab.name_offsets = calloc(max_files, sizeof(unsigned int));
    if (!_file_tab.name_offsets)
        return 1;
    _file_tab.names = malloc(NAME_CAP(max_files));
    if (!_file_tab.names)
        return 1;

    _file_tab.capacity  = max_files;
    _file_tab.name_head = 0;
    _file_tab.mutex = vyCreateMutex();
    return VY_SUCCESS;
}

void vyShutdownFileTab(void) {
    free(_file_tab.ids);
    free(_file_tab.names);
    free(_file_tab.name_offsets);
    vyDestroyMutex(_file_tab.mutex);
}


vy_file_id vyGetFileId(const char *path) {
    vy_text_span span;
    span.start  = path;
    span.length = (unsigned int)strlen(path);
    return vyGetFileIdFromSpan(span);
}

vy_file_id vyGetFileIdFromSpan(vy_text_span path) {
    /* Randomly choosen, aka finger smash keyboard */
    XXH64_hash_t seed = 15340978;
    vy_file_id fid    = (vy_file_id)XXH64(path.start, path.length, seed);
    if (fid == 0)
        fid = ~fid;
    return fid;
}

VY_DLLEXPORT vy_file_id vyAddFileFromSpan(vy_text_span path) {
    vy_file_id fid = vyGetFileIdFromSpan(path);

    if (!vyLockMutex(_file_tab.mutex)) {
        vyReportError("fio", "Failed to lock the guard mutex.");
        return 0;
    }

    /* Hash Insert */
    unsigned int i    = 0;
    unsigned int base = (unsigned int)(fid % _file_tab.capacity);
    while (i < _file_tab.capacity) {
        unsigned int at = (base + i) % _file_tab.capacity;
        if (_file_tab.ids[at] == 0) {
            /* Insert */
            unsigned int slen = (unsigned int)path.length + 1;
            if ((_file_tab.name_head + slen) >= NAME_CAP(_file_tab.capacity)) {
                /* Out of name storage */
                fid = 0;
                break;
            }
            memcpy(_file_tab.names + _file_tab.name_head, path.start, slen);
            _file_tab.name_offsets[at] = _file_tab.name_head;
            _file_tab.ids[at]          = fid;
            _file_tab.name_head += slen;
            break;
        } else if (_file_tab.ids[at] == fid) {
            break;
        }
        ++i;
    }

    /* Out of space */
    if (i == _file_tab.capacity)
        fid = 0;

    vyUnlockMutex(_file_tab.mutex);
    return fid;
}

VY_DLLEXPORT vy_file_id vyAddFile(const char *path) {
    vy_text_span span;
    span.start  = path;
    span.length = (unsigned int)strlen(path);
    return vyAddFileFromSpan(span);
}

VY_DLLEXPORT const char *vyGetFilePath(vy_file_id fid) {
    /* Hash Lookup */
    if (fid == 0)
        return NULL;

    if (!vyLockMutex(_file_tab.mutex)) {
        vyReportError("fio", "Failed to lock the guard mutex.");
        return 0;
    }
    const char *result = NULL;
    unsigned int i     = 0;
    unsigned int base  = (unsigned int)(fid % _file_tab.capacity);
    while (i < _file_tab.capacity) {
        unsigned int at = (base + i) % _file_tab.capacity;
        if (_file_tab.ids[at] == fid) {
            result = _file_tab.names + _file_tab.name_offsets[at];
            break;
        } else if (_file_tab.ids[at] == 0) {
            break;
        }
        ++i;
    }
    vyUnlockMutex(_file_tab.mutex);
    return result;
}
