#include <string.h>

#define VY_DONT_DEFINE_RENDERER_GLOBAL

#include "config.h"
#include "dynamic_libs.h"
#include "gfx.h"
#include "renderer_api.h"

/* Attributes are used to bind buffers (or textures) to symbolic values.
 * For example, an attribute might be bound to "CELL_GRID", which would be
 * replaced with the (at the time of the invoke) grid buffer of the current
 * world cell.
 */

vy_renderer_api g_renderer;
static vy_dynlib _renderer_lib;
static bool _renderer_loaded = false;

VY_CVAR_S(rt_Renderer,
          "Select the render backend. Available options: [vk], Default: vk",
          "vk");

extern bool
vyLoadShaders(const char **paths, vy_shader *shaders, unsigned int count);

#ifdef VY_STATIC_LIB
extern void VY_RENDERER_API_FN(RegisterCVars)(void);
extern vy_result VY_RENDERER_API_FN(Init)(const vy_renderer_init_info *);
extern void VY_RENDERER_API_FN(Shutdown)(void);
extern vy_gfx_pipeline_handle VY_RENDERER_API_FN(CompileComputePipeline)(
    const vy_compute_pipeline_info *);
extern vy_gfx_pipeline_handle VY_RENDERER_API_FN(CompileGraphicsPipeline)(
    const vy_graphics_pipeline_info *);
#endif

static bool LoadRenderer(void) {

#if !defined(VY_STATIC_LIB)
    #define RETRIEVE_SYMBOL(name, type)                                        \
        g_renderer.name = (type *)vyGetSymbol(_renderer_lib, "vyRen" #name);   \
        if (!g_renderer.name) {                                                \
            vyReportError(                                                     \
                "GFX",                                                         \
                "Unable to retrieve renderer function %s from backend %s",     \
                #name,                                                         \
                rt_Renderer.s);                                                \
        }

    if (strcmp(rt_Renderer.s, "vk") == 0) {
        _renderer_lib = vyOpenLib(VY_DLLNAME("vyvk"));
        if (!_renderer_lib) {
            vyReportError("GFX",
                          "Unable to load renderer backend: %s",
                          VY_DLLNAME("vyvk"));
            return false;
        }
        RETRIEVE_SYMBOL(RegisterCVars, vy_register_renderer_cvars_fn);
        RETRIEVE_SYMBOL(Init, vy_init_renderer_fn);
        RETRIEVE_SYMBOL(Shutdown, vy_shutdown_renderer_fn);
        RETRIEVE_SYMBOL(CompileComputePipeline, vy_compile_compute_pipeline_fn);
        RETRIEVE_SYMBOL(CompileGraphicsPipeline,
                        vy_compile_graphics_pipeline_fn);
    } else {
        vyReportError("GFX",
                      "Unsupported renderer backend: (%s) %s",
                      rt_Renderer.name,
                      rt_Renderer.s);
        return false;
    }
    #undef RETRIEVE_SYMBOL
#else
    g_renderer.RegisterCVars = &vyRenRegisterCVars;
    g_renderer.Init          = &vyRenInit;
    g_renderer.Shutdown      = &vyRenShutdown;
    g_renderer.CompileComputePipeline = &vyRenCompileComputePipeline;
    g_renderer.CompileGraphicsPipeline = &vyRenCompileGraphicsPipeline;
#endif
    return true;
}

VY_DLLEXPORT void vyRegisterRendererCVars(void) {
    if (!_renderer_loaded) {
        if (!LoadRenderer())
            return;
        _renderer_loaded = true;
    }
    g_renderer.RegisterCVars();
}

VY_DLLEXPORT bool vyInitGFX(vy_renderer_init_info *renderer_info) {
    if (!_renderer_loaded) {
        if (!LoadRenderer())
            return false;
        g_renderer.RegisterCVars();
    }

    if (g_renderer.Init(renderer_info) != VY_SUCCESS)
        return false;

    /* Init shader programs */
    const char *shader_files[] = {"shader/cell.shader"};
    vy_shader shaders[1];
    if (!vyLoadShaders(shader_files, shaders, 1))
        return false;
    return true;
}

VY_DLLEXPORT void vyShutdownGFX(void) {
    g_renderer.Shutdown();
}
